// SQL11.cpp : main project file.

using namespace System;
using namespace System::Threading;

/* Raima::Rdm is the system module that contains the ADO.NET
 * data provider for RDM */
using namespace Raima::Rdm;

/* EXAMPLE - SQL11 for C++ using ADO.NET */
/* This example shows multi-user access to a database using separate threads.
 * Two writer threads are created.  The first begins inserting values into
 * the table.  The second waits until the table contains more than 100 rows
 * and then begins deleting the oldest.  A reader thread is also created
 * that prints out the current oldest and newest entries every second.  Each
 * thread uses regular locking.
 * 
 * The following methods are illustrated:
 *     Raima::Rdm::RdmConnection:
 *         BeginTransaction() Method
 *         ChangeDatabase(string) Method
 *         Close() Method
 *         CreateCommand() Method
 *         Open() Method
 *         RdmConnection() Constructor
 *     Raima::Rdm::RdmCommand:
 *         CommandText Property
 *         ExecuteNonQuery() Method
 *         ExecuteReader() Method
 *         Parameters Property
 *     Raima::Rdm::RdmDataReader:
 *         Item[int] Property
 *         Read() Method
 *     Raima.Rdm.RdmParameterCollection
 *         Add(string, RdmType)
 *         Item[string] Property
 * 
 * They are fully document in http://docs.raima.com/rdm/.
 * 
 * For simplicity this example does not catch any thrown execptions
 * but good programming practices would dictate that this occur in
 * a full application.
 */

static void DropDatabase(
    RdmConnection ^conn,
    String        ^name)
{
    RdmCommand ^cmd = conn->CreateCommand();

    /* Attempt to drop this database.  If the database does not already exist
     * this will fail and we simply ignore the failure for this sample. */
    try
    {
        cmd->CommandText = "DROP DATABASE " + name;
        cmd->ExecuteNonQuery();

        cmd->CommandText = "COMMIT";
        cmd->ExecuteNonQuery();
    }
    catch (Exception ^)
    {
    }
}

static void CreateThisDatabase(
    RdmConnection   ^conn,
    String          ^name,
    array<String ^> ^cmds)
{
    RdmCommand ^cmd = conn->CreateCommand();

    cmd->CommandText = "CREATE DATABASE " + name;
    cmd->ExecuteNonQuery();

    for (int ii = 0; ii < cmds->Length; ii++)
    {
        cmd->CommandText = cmds[ii];
        cmd->ExecuteNonQuery();
    }

    cmd->CommandText = "COMMIT";
    cmd->ExecuteNonQuery();

    cmd->CommandText = "CLOSE " + name;
    cmd->ExecuteNonQuery();
}

static void CreateDatabases(
    RdmConnection ^conn)
{
    array<String ^> ^cmds = {
        "CREATE TABLE Table1 (id INTEGER PRIMARY KEY)"
    };

    /* Drop any existing copy of the databases. */
    DropDatabase(conn, "SQL11");

    /* Attempt to create the databases */
    CreateThisDatabase(conn, "SQL11", cmds);
}

static bool shutdown = false;

static void InserterThreadProc()
{
    /* Create a connection for this thread */
    RdmConnection ^conn = gcnew RdmConnection("host=local;database=SQL11;openmode=s");
    conn->Open();

    /* Create a command to do the insert */
    RdmCommand ^cmd = conn->CreateCommand();
    cmd->CommandText = "INSERT INTO Table1 VALUES :id";
    cmd->Parameters->Add("id", RdmType::Int32);

    /* Loop until a shutdown is requested */
    for (int ii = 0; !shutdown; ii++)
    {
        /* Insert the next value */
        cmd->Parameters["id"]->Value = ii;
        cmd->ExecuteNonQuery();
    }

    /* When a shutdown is requested, close the connection and exit */
    conn->Close();
}

static void DeleterThreadProc()
{
    /* Create a connection for this thread */
    RdmConnection ^conn = gcnew RdmConnection("host=local;database=SQL11;openmode=s");
    conn->Open();

    /* Create a command to check the number of rows, one to fine the 'smallest'
     * row, and one to delete */
    RdmCommand ^chkCmd = conn->CreateCommand();
    RdmCommand ^minCmd = conn->CreateCommand();
    RdmCommand ^delCmd = conn->CreateCommand();

    chkCmd->CommandText = "SELECT COUNT(id) FROM Table1";
    minCmd->CommandText = "SELECT MIN(id) FROM Table1";
    delCmd->CommandText = "DELETE FROM Table1 WHERE id = :minid";
    RdmParameter ^parm = delCmd->Parameters->Add("minid", RdmType::Int32);

    /* Loop until a shutdown is requested */
    while (!shutdown)
    {
        /* Query the number of rows in the table */
        RdmDataReader ^countRdr = chkCmd->ExecuteReader();
        if (countRdr->Read())
        {
            long count = (long) (Int64) countRdr[0];

            /* There should be only one row, so this closes the reader */
            countRdr->Read();

            if (count > 100)
            {
                /* If there at least 100 rows, get the smallest and delete it */
                RdmDataReader ^minRdr = minCmd->ExecuteReader();
                if (minRdr->Read())
                {
                    parm->Value = (int) minRdr[0];
                    delCmd->ExecuteNonQuery();

                    /* There should be only one row, so this closes the reader */
                    minRdr->Read();
                }
            }
        }
    }

    conn->Close();
}

static void ReaderThreadProc()
{
    /* Create a connection for this thread */
    RdmConnection ^conn = gcnew RdmConnection("host=local;database=SQL11;openmode=s");
    conn->Open();

    /* Create a command to do the read */
    RdmCommand ^cmd = conn->CreateCommand();
    cmd->CommandText = "SELECT MIN(id), MAX(id) FROM Table1";

    /* Loop until a shutdown is requested */
    while (!shutdown)
    {
        /* Wait one second */
        Thread::Sleep(1000);

        /* Query the mininum and maximum values */
        RdmDataReader ^rdr = cmd->ExecuteReader();
        if (rdr->Read())
        {
            Console::WriteLine("Current minimum = " + rdr[0] + " and current maximum = " + rdr[1]);
            rdr->Read();
        }
    }

    conn->Close();
}

int main(array<System::String ^> ^args)
{
    /* Create the connection with a simple connection string */
    RdmConnection ^conn = gcnew RdmConnection("host=local;openmode=s");
    conn->Open();

    CreateDatabases(conn);

    /* Create three threads one to insert, one to delete, and one to show the
     * current contents */
    Thread ^inserter = gcnew Thread(gcnew ThreadStart(InserterThreadProc));
    Thread ^deleter = gcnew Thread(gcnew ThreadStart(DeleterThreadProc));
    Thread ^reader = gcnew Thread(gcnew ThreadStart(ReaderThreadProc));

    Console::WriteLine("Starting threads");
    Console::WriteLine("   Press <Enter> when you are ready to terminate them");

    /* Start each of the threads */
    inserter->Start();
    deleter->Start();
    reader->Start();

    /* Wait for user to hit <Enter> */
    Console::ReadLine();

    /* Indicate to the threads that they should shut down */
    shutdown = true;

    /* Wait for each thread to finish */
    inserter->Join();
    deleter->Join();
    reader->Join();

    Console::WriteLine("All threads have terminated");

    conn->Close();

    Console::WriteLine("Hit <Enter> to continue");
    Console::ReadLine();
    return 0;
}
